@@ -27,14 +27,17 @@ import time
 import sys
 import os
 import ast
+from .util import copy_text_to_clipboard
+from .util import get_cmap_from_digitalizer_file
 
 import glob
 import re
 
-from .sxm import sxm_header
+from . import sxm
 
 import matplotlib.pyplot as plt
 from matplotlib import cm
+import matplotlib
 
 # BUG: interactive_colorplot.drag_bar appears to be broken in the MacOSX backend on matplotlib 3.5.1
 # To fix this issue, either use a different backend (such as Qt5Agg) or downgrade to matplotlib 3.4.3
@@ -533,7 +536,7 @@ class colorplot(interactive_colorplot.colorplot):
             plt.style.use('default')
 
         # Image data markers
-        self.img_data_points = {'filename': [],'V_s': [], 'V_g': []} # Contains filename, V_s and V_g for images that were taken and will be drawn as markers on the plot
+        self.img_data_points = {} # Keys should be the tuple (V_s, V_g) and values should be filenames
         self.marker_annot = self.ax.annotate("", xy=(0,0), xytext=(20,20),textcoords="offset points",
                     bbox=dict(boxstyle="round", fc="w"),
                     arrowprops=dict(arrowstyle="->"),
@@ -541,7 +544,28 @@ class colorplot(interactive_colorplot.colorplot):
         self.marker_annot.set_visible(False)
         self.img_data_scatter = self.ax.scatter([], [])
         plt.ion()
-        self.fig.canvas.mpl_connect("motion_notify_event", self.hover)
+        self.fig.canvas.mpl_connect("motion_notify_event", self.on_hover)
+        self.fig.canvas.mpl_connect("button_press_event", self.on_click)
+        self.real_cscale = 3.5
+        self.fft_cscale = 4
+        
+        #Set default colormap
+        path_to_this_file = os.path.dirname(__file__)
+        r_file = os.path.join(path_to_this_file, 'cmaps/w_r.csv')
+        g_file = os.path.join(path_to_this_file, 'cmaps/w_g.csv')
+        b_file = os.path.join(path_to_this_file, 'cmaps/w_b.csv')
+        self.sxm_cmap = get_cmap_from_digitalizer_file(r_file, g_file, b_file)
+
+    def get_header(self) -> dict:
+        '''
+        Returns the header of the first spectrum without the gate entry.
+        '''
+        header = self.spectra_list[0].header
+        try:
+            header.pop('Gate (V)')
+        except:
+            pass
+        return header
 
     @property
     def gate(self):
@@ -1055,6 +1079,19 @@ class colorplot(interactive_colorplot.colorplot):
         else:
             raise NotImplementedError(r"Unknown plotting library.")
 
+    def bias_and_gate_in_range(self, sample_bias, gate_voltage, tolerance=100):
+        '''
+        Returns True if sample_bias and gate_voltage are within the bounds of the gate sweep
+        '''
+        gate_min = np.amin(self.ylist)
+        gate_max = np.amax(self.ylist)
+        gate_tolerance = np.abs(gate_max - gate_min) / tolerance
+        bias_min = np.amin(self.xlist)
+        bias_max = np.amax(self.xlist)
+        bias_tolerance = np.abs(bias_max - bias_min) / tolerance
+
+        return (sample_bias > bias_min - bias_tolerance) & (sample_bias < bias_max + bias_tolerance) & (gate_voltage > gate_min - gate_tolerance) & (gate_voltage < gate_max + gate_tolerance)
+
     def add_img_data_marker(self, filename):
         '''
         Loads an sxm file "filename", grabs its sample bias and gate voltage, and adds 
@@ -1063,11 +1100,11 @@ class colorplot(interactive_colorplot.colorplot):
         If the gate voltage is not stored in the file, nothing is added to 
         self.image_data_marker.
         '''
-        header = sxm_header(filename)
+        header = sxm.sxm_header(filename)
 
         # Try getting gate voltage
         try:
-            gate_voltage = float(header[':Ext. VI 1>Gate voltage (V):'][0])
+            gate_voltage = round(float(header[':Ext. VI 1>Gate voltage (V):'][0]), 2)
         except:
             print("Warning: " + filename + " does not have the gate voltage stored in it")
             return
@@ -1076,16 +1113,10 @@ class colorplot(interactive_colorplot.colorplot):
             sample_biases = header['multipass biases']
             # Only add the marker if it's within the bounds of the spectrum
             for sample_bias in sample_biases:
-                if (sample_bias > np.amin(self.xlist)) & (sample_bias < np.amax(self.xlist)) & (gate_voltage > np.amin(self.ylist)) & (gate_voltage < np.amax(self.ylist)):
-                    self.img_data_points['filename'].append(filename)
-                    self.img_data_points['V_s'].append(sample_bias)
-                    self.img_data_points['V_g'].append(gate_voltage)
+                self.img_data_points[(round(sample_bias, 5), gate_voltage)] = filename
         else:
             sample_bias = float(header[':BIAS:'][0]) # If this throws an exception, then the header is probably fucked up
-            if (sample_bias > np.amin(self.xlist)) & (sample_bias < np.amax(self.xlist)) & (gate_voltage > np.amin(self.ylist)) & (gate_voltage < np.amax(self.ylist)):
-                self.img_data_points['filename'].append(filename)
-                self.img_data_points['V_s'].append(sample_bias)
-                self.img_data_points['V_g'].append(gate_voltage)
+            self.img_data_points[(round(sample_bias, 5), gate_voltage)] = filename
 
     def add_img_data_marker_manual(self, filename, sample_bias, gate_voltage):
         '''
@@ -1100,6 +1131,7 @@ class colorplot(interactive_colorplot.colorplot):
     #
     # Sometimes multiple images have the same (V_s, V_g) because the data was retaken
     # or because some images are incomplete. How should this be handled?
+    # Fixed using dicts
     def auto_add_img_data_markers(self, basename = None, start = 0, end = 99999, index_list = None):
         '''
         Automatically loops over all files in current working directory and adds image data markers using them.
@@ -1116,15 +1148,29 @@ class colorplot(interactive_colorplot.colorplot):
                             if int(match.group(1)) not in index_list: # Should check if index_list is a list, tuple, set, etc...
                                 continue
                         self.add_img_data_marker(filename)
+        self.plot_img_data_markers()
+        self.axes_reset()
 
     def auto_add_image_data_markers(self, *args, **kwargs):
+        '''
+        For name convenience
+        '''
         self.auto_add_img_data_markers(*args, **kwargs)
 
+    def add_img_data_markers_by_time(self, start_time, end_time):
+        '''
+        Adds files in current working directory if their last modified time is between 
+        start_time and end_time.
+        '''
+
+
     def plot_img_data_markers(self, s=100, color=(0, 0, 0, 1), zorder=1000, **kwargs):
         '''
         Plots the image data markers.
         '''
-        self.img_data_scatter = self.ax.scatter(self.img_data_points['V_s'], self.img_data_points['V_g'], s=s, color=color, zorder=zorder, **kwargs)
+        sample_biases = [bias for bias, _ in self.img_data_points.keys()]
+        gate_voltages = [gate for _, gate in self.img_data_points.keys()]
+        self.img_data_scatter = self.ax.scatter(sample_biases, gate_voltages, s=s, color=color, zorder=zorder, **kwargs)
         self.ax.zorder = 100
 
     def plot_image_data_markers(self, *args, **kwargs):
@@ -1134,12 +1180,79 @@ class colorplot(interactive_colorplot.colorplot):
         index = ind["ind"][0]
         pos = self.img_data_scatter.get_offsets()[index]
         self.marker_annot.xy = pos
-        text = self.img_data_points["filename"][index] + '\nVs = ' + str(self.img_data_points["V_s"][index]) + ' V' + '\nVg = ' + str(self.img_data_points["V_g"][index]) + ' V'
+        text = self.img_data_points[list(self.img_data_points.keys())[index]] + '\nVs = ' + str(list(self.img_data_points.keys())[index][0]*1000) + ' mV' + '\nVg = ' + str(list(self.img_data_points.keys())[index][1]) + ' V'
         self.marker_annot.set_text(text)
         self.marker_annot.get_bbox_patch().set_facecolor((0, 1, 0, 1))
         self.marker_annot.get_bbox_patch().set_alpha(0.6)
 
-    def hover(self, event):
+    def get_onenote_info_string(self) -> str:
+        '''
+        Returns an info string to paste into your notes
+        '''
+        filename = self.arg_list[0]
+        
+        header = self.get_header()
+
+        gate_range = f"{(round(np.amin(self.ylist), 2), round(np.amax(self.ylist), 2))} V"
+        gate_increment = f"{round(np.abs(self.ylist[1] - self.ylist[0]), 2)} V"
+
+        bias_range_float = (np.amin(self.xlist), np.amax(self.xlist))
+        if np.abs(bias_range_float[0]) < 1 or np.abs(bias_range_float[1]) < 1:
+            bias_range = f"{(round(bias_range_float[0]*1000, 2), round(bias_range_float[1]*1000, 2))} mV"
+        else:
+            bias_range = f"{(round(bias_range_float[0], 2), round(bias_range_float[1], 2))} V"
+
+        try:
+            setpoint_bias_float = float(header['Bias (V)'])
+            if np.abs(setpoint_bias_float) < 1:
+                setpoint_bias = f"{round(setpoint_bias_float * 1000, 2)} mV"
+            else:
+                setpoint_bias = f"{round(setpoint_bias_float, 2)} V"
+        except:
+            setpoint_bias = "Not recorded"
+        try:
+            setpoint_current_float = float(header['Setpoint current (pA)'])
+            if np.abs(setpoint_current_float) < 1e3:
+                setpoint_current = f"{round(setpoint_current_float, 2)} pA"
+            elif 1e3 < np.abs(setpoint_current_float) < 1e6:
+                setpoint_current = f"{round(setpoint_current_float / 1e3, 2)} nA"
+            else:
+                setpoint_current = f"{round(setpoint_current_float / 1e6, 2)} uA"
+        except:
+            setpoint_current = "Not recorded"
+        try:
+            lockin_amplitude_float = float(header['Lockin Amplitude'])
+            bias_calibration_factor = float(header['Bias Calibration Factor'])
+            if 0.3 < bias_calibration_factor < 0.7:
+                lockin_amplitude = f"{lockin_amplitude_float / 100 * 1000} mV"
+            elif bias_calibration_factor < 0.3:
+                lockin_amplitude = f"{lockin_amplitude_float} mV"
+            else:
+                lockin_amplitude = "Weird bias calibration value?"
+        except:
+            lockin_amplitude = "Not recorded"
+        try:
+            lockin_frequency = f"{header['Lockin Frequency']} Hz"
+        except:
+            lockin_frequency = "Not recorded"
+        try:
+            lockin_sensitivity = header['Lockin Sensitivity']
+        except:
+            lockin_sensitivity = "Not recorded"
+        try: 
+            lockin_time_constant = header['Lockin Time Constant']
+        except:
+            lockin_time_constant = "Not recorded"
+        
+        return f"{filename}\n\nGate range = {gate_range}\nGate increment = {gate_increment}\nBias range = {bias_range}\nSetpoint bias = {setpoint_bias}\nCurrent = {setpoint_current}\nLockin amplitude = {lockin_amplitude}\nLockin frequency = {lockin_frequency}\nLockin sensitivity = {lockin_sensitivity}\nLockin time constant = {lockin_time_constant}"
+        
+    def copy_onenote_info_string(self):
+        '''
+        Copies the string returned by get_onenote_info_string() onto the clipboard.
+        '''
+        copy_text_to_clipboard(self.get_onenote_info_string())
+
+    def on_hover(self, event):
         '''
         Handles mouse hover events over scatter plot points.
         '''
@@ -1154,6 +1267,52 @@ class colorplot(interactive_colorplot.colorplot):
                 if visible:
                     self.marker_annot.set_visible(False)
                     self.fig.canvas.draw_idle()
+
+    def on_click(self, event):
+        '''
+        Handles click events (namely the image data markers)
+        '''
+        if event.inaxes == self.ax:
+            cont, ind = self.img_data_scatter.contains(event)
+            if cont:
+                if sys.version_info[1] < 7:
+                    print("Error: Python version is below 3.7.")
+                    return
+                index = ind["ind"][0]
+                image_sxm = sxm.sxm(self.img_data_points[list(self.img_data_points.keys())[index]])
+                x_range = image_sxm.header['x_range (nm)']
+                y_range = image_sxm.header['y_range (nm)']
+                x_pixels = image_sxm.header['x_pixels']
+                y_pixels = image_sxm.header['y_pixels']
+                data = image_sxm.process_data(image_sxm.data["Z (m)"][0], process='subtract plane')
+                data_fft = np.abs(np.fft.fftshift(np.fft.fft2(data)))
+                
+                data_vmin = np.mean(data) - self.real_cscale*np.std(data)
+                data_vmax = np.mean(data) + self.real_cscale*np.std(data)
+                
+                fig, ax = plt.subplots(1, 2)
+                
+                copy_text_to_clipboard(image_sxm.get_onenote_info_string())
+
+                ax[0].imshow(data,
+                            cmap=self.sxm_cmap,
+                            vmin=data_vmin,
+                            vmax=data_vmax,
+                            origin='lower',
+                            extent=(0, x_range, 0, y_range)
+                )
+                ax[1].imshow(data_fft,
+                            cmap=self.sxm_cmap,
+                            vmin=0,
+                            vmax=self.fft_cscale*np.std(data_fft),
+                            origin='lower',
+                            extent = (-np.pi/(x_range/x_pixels), np.pi/(x_range/x_pixels), -np.pi/(y_range/y_pixels), np.pi/(y_range/y_pixels))
+                )
+                fig.canvas.draw()
+                plt.show()
+                
+
+
         
 
 def batch_load(basename, file_range = None, attribute_list = None, cache = None, constraint = None):
@@ -1190,7 +1349,7 @@ def batch_load(basename, file_range = None, attribute_list = None, cache = None,
                     continue
         return (spectrum_array, file_list)
 
-def parse_arguments(*spectra_arguments, **kwargs):
+def parse_arguments(*spectra_arguments, **kwargs) -> list[spectrum]:
 
     cache = kwargs['cache'] if ('cache' in kwargs) else None
     constraint = kwargs['constraint'] if ('constraint' in kwargs) else None
